Unit uThreads; {version 2.0 of Wil Barath's Thread unit for TP7.0}
               {see end of file for example and mission statement}
Interface      {(c)1995,1999 by Wil Barath.  see included docs}
{$a+,b-,d+,e-,g+,l-,o-,p-,q-,r-,s-,t-,v-,x+,y-}


{ Mission: to create a multithreading runtime for TP7.0 with:

  1 thread scheduling priorities

      threads are scheduled in descending priority, so you
    can group your threads strategicly if you like, and change
    the order at any time by changing priorities.

  2 optional preemptive scheduling

      preemptive scheduling means that threads will be scheduled
    when all higher-priority preemptive threads are 'waiting.'
    this allows for a loose sort of realtime schedule.

  3 simple synchronization

      threads can be signalled with 4 user-defined flags
    for synchronization or simple communication purposes.

  4 wait() for message/event

      threads (and the main program) can wait() on a pointer
    or longint to become valid/non-zero.  This allows for
    semaphores and preemptive scheduling, makes the scheduling
    process much faster vs entering and exiting user idle code.

  5 tiny memory footprint so this will be usable in Realmode space!

    using all features of the unit will increase your code size
    by 1157 bytes and your (global) data size by 24 bytes.

  todo / done:

        strategy and interface
        allocation of threads sorted by priority
        signal set/reset functions
        priority set/get functions
        thread authentication for all user functions

        scheduler actually working :)
        scheduler implements and respects flag settings
        scheduler converted to pure ASM
        wait() handled by scheduler
        preemptive threads
        scheduler removes TERM flagged threads
        thread(s) exit properly

        interface verbosely documented
        implementation code properly documented
        beta testing
        first public release

    x    message/event/mutex unit to compliment this one
    x    semaphore driven IO (file/modem/mouse)
    x    semaphore driven graphics unit
    x    GUI similar to my prev GUI unit based on above
    x    distributed threads  (wet dream ;)

}

Type
  pThreadPort=^Threadport;

  ThreadPort=Record {these should be considered as READ ONLY}
    Port:pointer;   {this is the port pointer we wait() on}
    Flags:byte;     {this is the flags of our thread}
    Priority:Byte;  {this is our priority}
  end;
  ThreadProc=Procedure(var mythread:threadport);
  {*** NOTE *** 'mythread' is given for convenience. It is READ ONLY}

Const
  tfwait=    $01;   {run if port isn't NIL }
  tfpreempt= $02;   {force all lower priority threads to wait}
  tfkill=    $04;   {thread should exit by itself }
  tfterm=    $08;   {thread will be removed before being scheduled again }
  tfsync1=   $10;   {thread should clear this when ready }
  tfsync2=   $20;   {thread should clear this when ready }
  tfsync3=   $40;   {thread should clear this when ready }
  tfsync4=   $80;   {thread should clear this when ready }

  minstack=  $010;   {stack size can be as small as 24 bytes and still
                     be able to call 'pass' for scheduling. However the
                     system may require more for interrupts etc.}

DefaultPriority:byte=100;  {valid priorities are from 0 to 127}
DefaultFlags:byte=0;       {default to no flags}
DefaultPort:Pointer=NIL;   {I'm leaving this entirely up to you.}
DefaultSize:Word=$080;     {enough to use writeln() conservatively}
NumThreads:Word=0;         {should be treated as READ-ONLY}

Function AddThread(proc:Threadproc):pThreadPort;
 {Create a thread with defaults for all fields and
  Insert it into the scheduling queue.
  returns NIL if unsuccessful}
Function DisposeThread(Var mythread:ThreadPort):Boolean;
 {removes the thread from the scheduling queue (if it was there)
  then deallocates it.  Returns false on NIL pointer}
Function CreateThread(Proc:ThreadProc;port:pointer;size:word;priority,flags:byte):pThreadPort;
 {allocate a thread and return pointer to it's message port}
Function InsertThread(VAR mythread:Threadport):Boolean;
 {insert thread into the scheduling queue
  *** will not let you insert it more than once!}
Function RemoveThread(VAR mythread:Threadport):Boolean;
 {remove thread from the scheduling queue
  *** will return true if the thread was in the queue}
Function SignalThread(VAR mythread:Threadport;sigflags:byte;setflag:boolean):Boolean;
 {returns true if the thread was in the queue}
Function SetThreadPri(VAR mythread:Threadport;newpri:byte):Boolean;
 {returns true if the thread was in the queue
  *** valid priorities are 0 to 127.  A priority of 0
  will suspend scheduling of the thread}
Function RestartThread(VAR mythread:ThreadPort):Boolean;
 {reinitializes a thread's stack; next pass will be treated like its first
  *** the thread must be reentrant.  ie. no assignment of typed constants}

Function GetThreadPri(VAR mythread:Threadport):Integer;
 {returns -1 if the thread wasn't in the queue}

Procedure Pass;
 {schedule threads.  I can't really tell you how to best use
  this, except to say that it should be in every idle loop
  in your program.  it should also be in any large loop that
  would otherwise lock up your program for too long.  Don't be
  afraid to throw it in a lot of places even in the same procedure
  as the overhead is really quite small for calling it.}

Procedure Wait(port:pointer);
 {wait until thread's port points to non zero value

  ***  note that this assigns the thread's port!
  ***  may also be called from MAIN}


Implementation

Type
pThread = ^Thread;

Thread = Record
  tPort:pointer;   {thread's message port *must* be the first field}
  tFlags:byte;     {thread's status flags}
  tPri:byte;       {thread's scheduling priority}
  tpNext:pThread;  {pointer to next thread in the list}
  tProc:ThreadProc;{thread's codespace}
  tSize:Word;      {size of the thread- stack and structure}
  tStack:Word;     {stack pointer of the thread}
end;

const                      {these *MUST* be NON-STACK variables!}
  ThreadList:pThread=NIL;  {we start with no threads}
  Current:pThread=NIL;     {and with no current thread}
  MAIN_SS:Word=0;          {place to store the main program stack}
  MAIN_SP:Word=0;          {pointer... if we want to get back there ;}

Function inQueue(p:pThread):Boolean;
var l:pthread;
begin
  l:=Threadlist;
  While (l<>NIL) and (l<>p) do l:=l^.tpnext;
  InQueue:=(l<>NIL);
end;

Procedure Insert(p:pThread);{link given thread}
var c:pthread;l:pointer;
begin
 l:=@threadlist;           {l remembers what to modify}
 c:=threadlist;            {c remembers the current thread}
 while ((c<>NIL) and (c^.tpri>p^.tpri)) do
 begin                     {search the thread list until we find}
   l:=@c^.tpNext;          {the end of the list, or a thread with}
   c:=c^.tpnext;           {a lower or equal priority}
 end;
 pointer(l^):=p;           {point previous thread to our one}
 p^.tpnext:=c;             {point our thread to the next one}
 inc(numthreads);          {adjust our thread counter}
end;

Procedure Remove(p:pthread); {unlink given thread}
var c:pthread;n:pointer;
begin
  n:=@threadlist;          {n remembers what to modify}
  c:=threadlist;           {c remembers the current thread}
  while c<>p do            {find the thread}
  begin
    n:=@c^.tpnext;
    c:=c^.tpnext;
  end;
  pointer(n^):=p^.tpnext;  {point thread before last to next one}
  dec(numthreads);         {adjust our thread counter}
end;

{***************************************************************************}
{*                                                                         *}
{*  This is where the user-called functions are.  Most of these are just   *}
{*  stubs that do some safety checks and a little typecasting before       *}
{*  calling off to a feature already implemented for another reason.       *}
{*                                                                         *}
{***************************************************************************}

Function CreateThread(Proc:Threadproc;port:pointer;size:Word;priority,flags:byte):pThreadPort;
var p:pthread;
begin
  p:=nil;                           {what to return on failure}
  if size>=minstack then            {do not allow illegal stack size}
  begin
    getmem(p,size+sizeof(thread));  {try to get the memory}
    if p<>NIL then                  {if we got it, initialize it}
    begin
      with p^ do                    {initialize the fields}
      begin
        tProc:=Proc;
        tPort:=Port;
        tSize:=size+Sizeof(thread);
        tPri:=Priority;
        tFlags:=Flags;
        tStack:=0;                  {this lets pass know it's new}
      end;
    end;
  end;
  createThread:=pthreadport(p);              {return pointer to the port, or NIL}
end;

Function InsertThread(var mythread:threadport):Boolean;
begin
  if (@mythread=NIL) or InQueue(@mythread)then insertthread:=false
  else     {don't insert it if NIL or if it's already in the queue!}
  begin
    insert(@mythread);
    insertthread:=true;
  end
end;

Function RemoveThread(var mythread:Threadport):Boolean;
begin
  if InQueue(@mythread) then                {only remove if it's there}
  begin
    remove(@mythread);
    removethread:=true;
  end
  else removethread:=false;
end;

Function AddThread(proc:Threadproc):pThreadPort;
var p:pthreadport;
begin                                       {create thread with all defaults}
  p:=createthread(proc,defaultport,defaultsize,defaultpriority,defaultflags);
  if p<>NIL then insert(pthread(p));        {if successful, insert it}
  AddThread:=p;
end;

Function DisposeThread(var mythread:ThreadPort):Boolean;
var result:boolean;
begin
  if (@mythread <>NIL) then  {make sure it's valid}
  begin
    RemoveThread(mythread);  {make sure it isn't in the queue!}
    freemem(@mythread,pthread(@mythread)^.tsize);
    disposethread:=true;
  end
  else disposethread:=false;
end;

Function SignalThread(var mythread:Threadport;sigflags:byte;setflag:boolean):Boolean;
begin
  If InQueue(@mythread)then with mythread do
  begin
    SignalThread:=true;
    if setflag then flags:=flags OR sigflags
    else flags:=flags and (not sigflags);
  end
  else SignalThread:=false;
end;

Function SetThreadPri(var mythread:Threadport;newpri:byte):Boolean;
begin
  If InQueue(@mythread)then with mythread do
  begin
    priority:=newpri;
    remove(@mythread);
    insert(@mythread);
    SetthreadPri:=true;
  end
  else SetThreadPri:=false;
end;

Function GetThreadPri(var mythread:Threadport):Integer;
begin
  If InQueue(@mythread)then GetThreadPri:=mythread.priority
  else GetThreadPri:=-1;
end;

Function RestartThread(VAR mythread:ThreadPort):Boolean;
begin
  pthread(@mythread)^.tstack:=0;
end;

Procedure Wait(port:pointer);assembler;
asm
  les di,current                     {find current thread}
  mov cx,es                          {see if it's main}
  jcxz @maincaller                   {yes- do loop for main}
  mov ax,[port].word                 {otherwise assign the port}
  mov es:[di].thread.tport.word,ax
  mov ax,[port+2].word
  mov es:[di+2].thread.tport.word,ax
  or  es:[di].thread.tflags,tfwait   {set the wait flag}
  call pass                          {and pass off}
  jmp  @done

@maincaller:
  call pass                          {always pass once}
  les di,port                        {test the port}
  mov ax,es:[di].word
  or  ax,es:[di+2].word
  jz  @maincaller                    {if zero, do it again}
@done:
end;

procedure ThreadExit; {remove the current thread}
var t:pthread;
begin
  t:=current;              {remember where it is}
  current:=current^.tpnext;{advance scheduler to next thread}
  remove(t);               {unlink it}
  freemem(t,t^.tsize);     {deallocate it}
end;

{***************************************************************************}
{*                                                                         *}
{*  This is where all the action is.  Here we unlink the stack of the      *}
{*  main program or the last scheduled thread and switch to the stackspace *}
{*  of the next scheduled thread, or exit back on the main program stack   *}
{*                                                                         *}
{***************************************************************************}

procedure pass;assembler;
const preempt:byte=0;
asm
        cli                             {disable interrupts}
        les     di,current              {test our caller}
        mov     cx,es
        jcxz    @maincaller
        push    bp                      {called from thread}
        mov     es:[di].thread.tstack,sp{save the thread's stack}
        les     di,es:[di].thread.tpnext{move to next thread}
        mov     cx,es

@updatecurrent:
        mov     [current].word,di       {save current pointer}
        mov     [current+2].word,es

@schedule:
        jcxz    @mainexit               {exit to MAIN if NIL}
        mov     ax,es:[di].thread.tflags.word  {get flags and priority}
        test    al,tfterm               {see if thread is marked for death}
        jnz     @terminatethread        {if so, then terminate it}
        cmp     ah,preempt              {test our priority}
        jb      @clearexit              {remaining threads are skipped}
        test    al,tfwait               {see if thread is waiting}
        jz      @nowait
        les     si,es:[di].thread.tport {test port^}
        mov     bx,es:[si].word
        or      bx,es:[si+2].word
        mov     es,cx
        jz      @nextthread             {if NIL then skip}
        and     es:[di].thread.tflags,not tfwait{clear wait flag}

@nowait:
        test    al,tfpreempt            {see if thread is preemptive}
        jz      @runthread              {no- just run it}
        mov     preempt,ah              {yes- set preemptive level}

@runthread:
        mov     ss,cx                   {otherwise set up stack}
        mov     cx,es:[di].thread.tstack{initialized already?}
        jcxz    @initthread             {no- go do it}
        mov     sp,cx                   {yes- set up stack}
        pop     bp
        sti                             {restore interrupts}
        retf                            {exit back into thread}

@clearexit:
        xor     bx,bx                   {exiting to main}
        mov     [current].word,bx       {set current to NIL}
        mov     [current+2].word,bx     {}

@mainexit:
        mov     ss,MAIN_SS              {restore the MAIN stack}
        mov     sp,MAIN_SP
        pop     bp
        sti                             {restore interrupts}
        retf                            {exit back to main}

@nextthread:
        les     di,es:[di].thread.tpnext{get next thread}
        mov     cx,es                   {prepare cx for NIL test}
        jmp     @updatecurrent

@maincaller:
        mov     preempt,1               {set preemptive level to 1}
        les     di,threadlist           {get pointer to first thread}
        mov     cx,es                   {see if it's good}
        jcxz    @done                   {no threads- exit 'fast'}
        push    bp                      {otherwise save main stack}
        mov     MAIN_SS,ss              {then make the thread current}
        mov     MAIN_SP,sp              {and follow through to schedule}
        jmp     @updatecurrent

@initthread:
        mov     sp,di                   {base of thread is base of stack}
        add     sp,es:[di].thread.tsize {add size of thread and}
        dec     sp                      {subtract one = top of stack}
        push    [current+2].word        {push pointer to the threadport}
        push    [current].word
        sti                             {restore interrupts}
        call    es:[di].thread.tproc    {call the thread}

@terminatethread:
        call    threadexit              {when it exits, it dies}
        les     di,current              {try the next thread}
        mov     cx,es
        jmp     @schedule
@done:                                  {used only for 'fast' exit}
        sti                             {restore interrupts}
end;
end. {Unit Threads}



{***************************************************************************}
{*                                                                         *}
{* This is a little example that serves as a minimalist proof-of-concept   *}
{*                                                                         *}
{***************************************************************************}


Program testthreads;
  {test the basic features of the thread unit}

uses hardware,uthreads; {'crt' will crash on pentiums without a patch}
var               {you can use my 'hardware' unit (available from SWAG) instead}
  l:word;
  p:pthreadport;

const keycode:longint=0;

procedure Keyboard(var threadstuff:threadport);far;
begin
  repeat
    writeln('Lowpriority keyboard reader');
    if (keycode=0) and keypressed then keycode:=longint(readkey);
    pass;
  until false;  {never exit}

end;

procedure preempt(var threadstuff:threadport);far;
var l:word;
begin
  setthreadpri(threadstuff,101);
  signalthread(threadstuff,tfpreempt,true);
  Writeln('Preemptive code running at priority 101');
  repeat
    writeln('debug: hit a key to continue');
    readkey;
    for l:=1 to 3 do
    begin
      writeln('preemptive level 101, pass',l);
      pass;
    end;
    writeln('debug: hit a key to continue');
    writeln('then "t" to terminate hipri, or "esc" to exit');
    readkey;
    writeln('preemptive level 101 waiting on keyboard');
    wait(@keycode);
    if keycode=ord('t') then
    begin
      if signalthread(p^,tfterm,true)
      then writeln('tee-hee... hipri is a GONER!')
      else writeln('huh?  where''d hipri go?');
    end;
    if keycode<>27 then keycode:=0;
  until false;
end;

procedure hipriority(var threadstuff:threadport);far;
var l:word;
begin
  setthreadpri(threadstuff,102);
  repeat
    writeln('hipriority 102');
    pass;
  until false;
end;

begin
  Writeln('Thread unit test program');

  addthread(keyboard);
  addthread(preempt);
  p:=addthread(hipriority);
  repeat
    writeln('in main code');
    wait(@keycode);
  until keycode=27;

  writeln('Main code exiting');
end.
